// SD Interface v0.1
//
// by J.H.NOCK, November 2021
//
// Connects the SD modules to the rest of the GPU and manages the
// 4KB dual-port M9K block RAM read/write buffer.
//
// TODO: Get reads working with direction controls.
//     * Handle direction control for SD_CMD in this module.
// TODO: Handle write requests to the SD card.

module SDInterface #(

    parameter int          CLK_DIV     = 3,      // clock divider
    parameter int          BUFFER_ADDR = 'h5000  // default DDR3 buffer location

)(

    input  logic           CLK,            // 100MHz system clock
    input  logic           RESET,          // reset active HIGH
    // interface <-> Bridgette
    input  logic           RD_REQ,         // HIGH for read request
    input  logic   [ 31:0] SECTOR,         // sector number to read/write
    output logic           RD_RDY,         // HIGH when read is complete from SD card
    output logic           SD_BUSY,        // HIGH when interface is busy
    output logic   [  3:0] SIDSTATE,       // current state machine value
    output logic   [  1:0] CARDTYPE,       // SD card type
    output logic           BUSY,           // HIGH when SDInterface is busy
    output logic   [  7:0] STATUS,         // aggregated SD status byte of above 4 outputs
    // SD phy connections
    inout  logic   [  3:0] SD_DATA,        // data from SDReader.sv
    inout                  SD_CMD,         // CMD signal to SDReader.sv
    output logic           SD_CLK,         // clock signal to SD card
    output logic           SD_CMD_DIR,     // HIGH = TO SD card, LOW = FROM SD card
    output logic           SD_D0_DIR,      // HIGH = TO SD card, LOW = FROM SD card
    output logic           SD_D123_DIR,    // HIGH = TO SD card, LOW = FROM SD card
    output logic           SD_SEL,         // SD select
    // DDR3 connections
    //    input  (DDR3 -> SD WR ops)
    input  logic           DDR3_busy,      // HIGH when DDR3 is busy
    input  logic   [127:0] DDR3_rd_data,   // read data from DDR3
    input  logic           DDR3_rd_rdy,    // data from DDR3 is ready
    //    output (SD -> DDR3 RD ops)
    output logic    [31:0] DDR3_addr_o,
    output logic           DDR3_ena,       // Flag HIGH for 1 CMD_CLK when sending a DDR3 command
    output logic           DDR3_wr_ena,    // HIGH signals write request to DDR3 Controller
    output logic   [127:0] DDR3_wr_data,   // 128-bit data bus
    output logic    [15:0] DDR3_wr_mask    // Write data enable mask

);

// data IO for read and write data, to connect to buffer
wire          cache_wren ;
wire          rd_req     ; // HIGH when not BUSY and RD_REQ is made by Bridgette
wire    [1:0] cardtype   ;

logic         cache_full ; // HIGH when 16-byte cache is full
logic         rEna       ; // HIGH whilst rData is being sent
logic [  7:0] SD_rData   ; // Data read from cache
logic [  8:0] buf_ptr    ; // 512-byte buffer address pointer
logic [  3:0] data_ptr   ; // pointer to current byte in cache
logic [127:0] data_cache ; // temporary 16-byte storage to be written to DDR3 or SD

//                         1        2         4        1
assign STATUS       = '{ RD_RDY, cardtype, SIDSTATE, SD_BUSY } ;
assign CARDTYPE     = cardtype                 ;
assign rd_req       = BUSY ? 1'b0 : RD_REQ     ; // only pass rd_req if interface isn't busy
// DDR3 Write Request
assign DDR3_wr_ena  = cache_full && !DDR3_busy ; // Set the write enable.
assign DDR3_wr_data = data_cache               ; // Send write data.
assign DDR3_wr_mask = 16'hFFFF                 ; // Write enable for the byte.
// DDR3 transaction controls
assign DDR3_addr_o  = BUFFER_ADDR + buf_ptr    ; // DDR3 write address.
assign DDR3_ena     = DDR3_wr_ena              ; // Write full 128-bits.

// For more detail, see SDReader.sv
SDReader #(

    .CLK_DIV         ( CLK_DIV     )  // because clk=100MHz, CLK_DIV is set to 2 - see SDReader.sv for detail

) SDReader_inst(

    .clk             ( CLK         ),
    .rst_n           ( !RESET      ), // rst_n is active low, so RESET must be inverted
    
    // signals connect to SD bus
    .sdclk           ( SD_CLK      ),
    .sdcmd           ( SD_CMD      ),
    .sddat           ( SD_DATA     ),
    
    // bus direction controls
    .SD_CMD_DIR      ( SD_CMD_DIR  ), // HIGH = TO SD card, LOW = FROM SD card
    .SD_D0_DIR       ( SD_D0_DIR   ), // HIGH = TO SD card, LOW = FROM SD card
    .SD_D123_DIR     ( SD_D123_DIR ), // HIGH = TO SD card, LOW = FROM SD card
    .SD_SEL          ( SD_SEL      ), // SD socket select

    // status and information
    .card_type       ( cardtype    ), // 0=Unknown, 1=SDv1.1 , 2=SDv2 , 3=SDHCv2
    .card_stat       ( SIDSTATE    ), // current state of SDReader's state machine
    
    // user read sector command interface
    .rstart          ( rd_req      ), // rstart HIGH starts read operation
    .rsector_no      ( SECTOR      ), // target sector to read in SDcard
    .rbusy           ( SD_BUSY     ), // signals read is ongoing or complete
    .rdone           ( RD_RDY      ), // signals read is complete
    
    // sector data output interface
    .outreq          ( cache_wren  ), // HIGH whilst data is received from SD card (for each byte?)
    .outaddr         ( readAddr    ), // cache address to be written to
    .outbyte         ( SD_rData    )  // data read from SD card

);

always @( posedge CLK ) begin

    if ( RESET ) begin

        buf_ptr    <= 'hFF0  ; // start at 0xFF0 so when first read completes, buf_ptr = 0 for first memory write
        cache_full <= 1'b0   ;
        data_cache <= 128'b0 ;
        data_ptr   <= 4'b0   ;
        BUSY       <= 1'b0   ;

    end else if ( rd_req ) begin // new read request received

        buf_ptr    <= 'hFF0  ;
        BUSY       <= 1'b1   ;

    end else if ( !RD_RDY && cache_full && !DDR3_busy ) begin // 16-byte cache is full

        cache_full <= 1'b0   ;

    end else if ( RD_RDY ) begin // End of SD Read; reset pointers and data bus
    
        buf_ptr    <= 'hFF0  ;
        cache_full <= 1'b0   ;
        data_cache <= 128'b0 ;
        data_ptr   <= 4'b0   ;
        BUSY       <= 1'b0   ;

    end else if ( cache_wren ) begin // byte received from SDReader

        case ( data_ptr )

            0  : data_cache[127:120] <= SD_rData ; // 127-120
            1  : data_cache[119:112] <= SD_rData ; // 119-112
            2  : data_cache[111:104] <= SD_rData ; // 111-104
            3  : data_cache[103: 96] <= SD_rData ; // 103-96
            4  : data_cache[ 95: 88] <= SD_rData ; // 95-88
            5  : data_cache[ 87: 80] <= SD_rData ; // 87-80
            6  : data_cache[ 79: 72] <= SD_rData ; // 79-72
            7  : data_cache[ 71: 64] <= SD_rData ; // 71-64
            8  : data_cache[ 63: 56] <= SD_rData ; // 63-56
            9  : data_cache[ 55: 48] <= SD_rData ; // 55-48
            10 : data_cache[ 47: 40] <= SD_rData ; // 47-40
            11 : data_cache[ 39: 32] <= SD_rData ; // 39-32
            12 : data_cache[ 31: 24] <= SD_rData ; // 31-24
            13 : data_cache[ 23: 16] <= SD_rData ; // 23-16
            14 : data_cache[ 15:  8] <= SD_rData ; // 15-8
            15 : begin
                 data_cache[  7:  0] <= SD_rData ; // 7-0
                 cache_full          <= 1'b1     ;
            end

        endcase
        buf_ptr  <= buf_ptr  + 1'b1 ;
        data_ptr <= data_ptr + 1'b1 ;

    end

end

endmodule
